package dicewars;

import dicewars.file.FileService;
import dicewars.model.Country;
import dicewars.model.DiceWars;
import dicewars.model.Hex;
import dicewars.model.Player;
import dicewars.util.Directions;

import dicewars.util.Point;
import javafx.animation.Animation;
import javafx.animation.ParallelTransition;
import javafx.animation.RotateTransition;
import javafx.application.Application;
import javafx.event.Event;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.*;
import javafx.scene.paint.Color;

import javafx.stage.Stage;
import javafx.util.Duration;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;

import static java.lang.Math.sqrt;
import static java.lang.System.exit;
import static java.lang.System.out;

/*
 *  JavaFX GUI and event handling for the Dice Wars Game
 *
 *
 */
public class DiceWarsGUI extends Application {

    private DiceWars diceWars;                  // The main model, created at init()
    private Country country;                    // Any selected country
    private Country from;                       // Selected country to attack from

    private ParallelTransition dicesAnimation;  // Fancy GUI (it works, nothing to do);

    // ---- Game handling --------------------

    void selectCountry(MouseEvent me) {

        // Return if animation running
        if (dicesAnimation != null && dicesAnimation.getStatus() == Animation.Status.RUNNING) {
            return;
        }

        // This is a pure mathematical mapping (possibly outside the grid)
        Hex hex = new Hex(new Point(me.getX(), me.getY()), hexRadius);

        // Is there a Country for the selected Hex
        country = diceWars.getCountryFor(hex);
        if (country == null) {
            return;
        }

        // TODO The selection algorithm
        // Also call model and renderNNN() methods and dices animation in the flow

    }

    private void next(MouseEvent me) {

        // TODO update model

        renderDices(0, 0);
        renderCountries();
        renderPlayers();
        renderAttackerDefender(defaultDefender);
    }

    /*
       This is how to use JavaFX built in animations (no explicit Timer)

       dicesAnimation = getDicesAnimation();
       dicesAnimation.setOnFinished(this::animationFinished);
       dicesAnimation.play();

     */
    // Run this when the animation is over i.e. setOnFinish()
    private void animationFinished(Event e) {
        from.setSelected(false);
        country.setSelected(false);
        renderCountry(from);
        renderCountry(country);
        renderPlayers();
        from = null;
    }

    // --- Animations -----------------------

    // Animate the dices (i.e. rotate two labels)
    private ParallelTransition getDicesAnimation() {
        RotateTransition ar =
                new RotateTransition(Duration.millis(50), attacker);
        ar.setByAngle(180f);
        ar.setCycleCount(10);
        ar.setAutoReverse(true);
        RotateTransition dr =
                new RotateTransition(Duration.millis(50), defender);
        dr.setByAngle(180f);
        dr.setCycleCount(10);
        dr.setAutoReverse(true);
        return new ParallelTransition(ar, dr);
    }

    // -------- Initialization ---------------------

    // Called by FX first of all, here we do the main setup
    // NOTE: This method is possible to debug, no graphics is up yet!
    @Override
    public void init() {
        // Hard coded for now
        Player[] ps = {new Player("Pelle", Color.LIGHTGREEN), new Player("Olle", Color.LIGHTBLUE),
                new Player("Fia", Color.YELLOW), new Player("Lisa", Color.ORANGE)};
        List<Player> players = Arrays.asList(ps);

        // Get the hex grid
        Map<Color, List<Hex>> regions = null;
        try {
            Path path = Paths.get("src/dicewars/grids/grid1.txt");
            regions = FileService.readHexGrid(path); // Hard coded for now
        } catch (IOException e) {
            e.printStackTrace();  // Later show dialog
            exit(0);
        }

        List<Country> countries = new ArrayList<>();

        // TODO construct model


        fixLayout(regions); // Must have this last (already implemented)
    }


    // ---- JavaFX below this, nothing to do for you ---------

    private final int height = 600;
    private final int width = 600;
    private double margin;                             // From pane border to hex grid
    private double hexRadius;                             // For all hexes (center to some corner)
    private final Map<Player, Label> playerLabel = new HashMap<>();  // Matching player with label
    private final Color selected = Color.GRAY;             // When country selected

    private final String redBorder = "-fx-border-color: red; -fx-border-width: 3";
    private final String blackBorder = "-fx-border-color: black; -fx-border-width: 1";
    private final String flowPane = "-fx-border-color: black; -fx-padding: 10; -fx-hgap: 10; -fx-alignment: center";
    private final String defaultDefender = "lightgray";
    private Label attacker;
    private Label defender;
    private GraphicsContext gc;
    Button btnNext = new Button("Next");

    private void fixLayout(Map<Color, List<Hex>> regions) {
        int nHex = 0;
        for (List<Hex> hs : regions.values()) {
            nHex += hs.size();
        }
        hexRadius = width / sqrt(nHex) / 2;
        margin = hexRadius;
    }

    @Override
    public void start(Stage primaryStage) throws Exception {

        BorderPane root = new BorderPane();
        root.setOnMouseClicked(this::selectCountry);

        VBox south = new VBox();
        FlowPane dicesPanel = createDicesPanel();
        FlowPane playersPanel = createPlayersPanel();
        south.getChildren().addAll(dicesPanel, playersPanel);
        root.setBottom(south);

        Canvas canvas = new Canvas(width, height);
        gc = canvas.getGraphicsContext2D();
        root.setCenter(canvas);

        renderCountries();
        renderAttackerDefender(defaultDefender);
        renderPlayers();

        Scene scene = new Scene(root);
        primaryStage.setScene(scene);
        primaryStage.setTitle("Dice Wars");
        primaryStage.show();
    }

    private FlowPane createPlayersPanel() {
        FlowPane fp = new FlowPane();
        fp.setStyle(flowPane);
        for (Player p : diceWars.getPlayers()) {
            Label l = new Label(p.getName());
            l.setStyle("-fx-padding: 15; -fx-background-color:" + colorToWeb(p.getColor()));
            fp.getChildren().add(l);
            playerLabel.put(p, l);
        }
        //Button next = new Button("Next");
        btnNext.setStyle("-fx-padding: 10 40 10 40; -fx-background-color: gray;");
        btnNext.setOnMouseClicked(this::next);
        fp.getChildren().add(btnNext);
        return fp;
    }

    private FlowPane createDicesPanel() {
        FlowPane dp = new FlowPane();
        dp.setStyle(flowPane);
        attacker = new Label("0");
        defender = new Label("0");
        dp.getChildren().addAll(attacker, defender);
        return dp;
    }

    // ----------- Rendering ------------------

    private void renderPlayers() {
        diceWars.getPlayers().forEach(p -> {
            Label l = playerLabel.get(p);
            if (p == diceWars.getActual()) {
                l.setStyle(l.getStyle() + ";" + redBorder);
            } else {
                l.setStyle(l.getStyle() + ";" + blackBorder);
            }
        });

    }

    private void renderAttackerDefender(String colorName) {
        attacker.setStyle("-fx-padding: 15; -fx-background-color:" + colorToWeb(diceWars.getActual().getColor()));
        defender.setStyle("-fx-padding: 15; -fx-background-color:" + colorName);
    }

    private void renderCountries() {
        diceWars.getCountries().forEach(this::renderCountry);
    }

    private void renderCountry(Country c) {
        double[] xs = new double[6];
        double[] ys = new double[6];

        for (Hex h : c.getHexes()) {
            List<Point> corners = h.getCorners(hexRadius);
            getXsAndYs(xs, ys, corners);
            if (c.isSelected()) {
                gc.setFill(selected);
            } else {
                gc.setFill(c.getOwner().getColor());
            }
            gc.fillPolygon(xs, ys, xs.length);

            // Render borders
            List<Directions> borders = c.getBorders(h);
            if (borders != null) {
                for (Directions d : borders) {
                    int i = d.ordinal();           // NOTE:  Depends on ordering of enum Directions!
                    gc.setStroke(Color.BLACK);
                    gc.setLineWidth(5);
                    gc.strokeLine(xs[i], ys[i], xs[(i + 1) % 6], ys[(i + 1) % 6]);
                }

            }
            renderCountryDice(c);
        }
    }

    // The values shown in each Country
    private void renderCountryDice(Country c) {
        Hex hex = c.getHexes().get(0);
        Point p = hex.toPoint(hexRadius);
        gc.setFill(Color.BLACK);
        gc.fillText(String.valueOf(c.getNDices()), p.x + margin, p.y + margin);

    }

    private void renderDices(int attackerResult, int defenderResult) {
        attacker.setText(String.valueOf(attackerResult));
        defender.setText(String.valueOf(defenderResult));
    }

    // Conversion from list of Points to parallel arrays of x and y -coordinates
    private void getXsAndYs(double[] xs, double[] ys, List<Point> pts) {
        int i = 0;
        for (Point p : pts) {
            xs[i] = p.x + margin;
            ys[i] = p.y + margin;
            i++;
        }
    }

    // Convert to web color codes for use with CSS
    private String colorToWeb(Color color) {
        return "#" + Integer.toHexString(color.hashCode());
    }

    public static void main(String[] args) {
        launch();
    }

}
